from datetime import timedelta

INNER_THOUGHT_TAG = "inner_thought_monologue"
SYSTEM_INSTRUCTION_LABEL = "*Elroy System Instruction*"

UNKNOWN = "Unknown"
MEMORY_TITLE_EXAMPLES = """
Examples of good and bad memory titles are below. Note, the BETTER examples, some titles have been split into two.:

BAD:
- [User Name]'s project progress and personal goals: 'Personal goals' is too vague, and the title describes two different topics.

BETTER:
- [User Name]'s project on building a treehouse: More specific, and describes a single topic.
- [User Name]'s goal to be more thoughtful in conversation: Describes a specific goal.

BAD:
- [User Name]'s weekend plans: 'Weekend plans' is too vague, and dates must be referenced in ISO 8601 format.

BETTER:
- [User Name]'s plan to attend a concert on 2022-02-11: More specific, and includes a specific date.

BAD:
- [User Name]'s preferred name and well being: Two different topics, and 'well being' is too vague.

BETTER:
- [User Name]'s preferred name: Describes a specific topic.
- [User Name]'s feeling of rejuvenation after rest: Describes a specific topic.
"""

CLI_USER_ID = 1

### Model parameters ###

# TODO: make this dynamic
EMBEDDING_SIZE = 1536

DEFAULT_CHAT_MODEL_NAME = "gpt-4o"
DEFAULT_CONTEXT_WINDOW_LIMIT = 16384
DEFAULT_EMBEDDING_MODEL_NAME = "text-embedding-3-small"

MIN_CONVO_AGE_FOR_GREETING = timedelta(minutes=10)

L2_MEMORY_RELEVANCE_DISTANCE_THRESHOLD = 1.24

L2_MEMORY_CONSOLIDATION_DISTANCE_THRESHOLD = 0.65

INITIAL_REFRESH_WAIT_SECONDS = 30


RESULT_SET_LIMIT_COUNT = 5

MEMORY_WORD_COUNT_LIMIT = 300
DEFAULT_ASSISTANT_COLOR = "#77DFD8"
DEFAULT_INPUT_COLOR = "#FFE377"
DEFAULT_SYSTEM_MESSAGE_COLOR = "#9ACD32"
DEFAULT_WARNING_COLOR = "yellow"
DEFAULT_INTERNAL_THOUGHT_COLOR = "#708090"

REPO_LINK = "https://github.com/elroy-bot/elroy/issues"


class MissingAssistantToolCallError(Exception):
    pass


class MissingToolCallMessageError(Exception):
    pass


class MissingSystemInstructError(Exception):
    pass


class MisplacedSystemInstructError(Exception):
    pass



import os
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import timedelta
from typing import Generator, Generic, Optional, TypeVar

from litellm import (
    anthropic_models,
    open_ai_chat_completion_models,
    open_ai_embedding_models,
)
from sqlalchemy import NullPool, create_engine
from sqlmodel import Session

from ..io.base import ElroyIO

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


@dataclass
class ChatModel:
    model: str
    api_key: Optional[str]
    ensure_alternating_roles: (
        bool  # Whether to ensure that the first message is system message, and thereafter alternating between user and assistant.
    )
    api_base: Optional[str] = None
    organization: Optional[str] = None


@dataclass
class EmbeddingModel:
    model: str
    embedding_size: int
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    organization: Optional[str] = None


@dataclass
class ElroyConfig:
    postgres_url: str
    context_window_token_limit: int
    context_refresh_token_trigger_limit: int  # how many tokens we reach before triggering refresh
    context_refresh_token_target: int  # how many tokens we aim to have after refresh
    max_in_context_message_age_seconds: int  # max age of a message to keep in context
    context_refresh_interval_seconds: int  # how often to refresh system message and compress context messages
    chat_model: ChatModel
    embedding_model: EmbeddingModel
    fail_fast: bool  # Whether to emit more verbose logging and fail faster on errors rather than attempting to recover


def get_config(
    postgres_url: str,
    chat_model_name: str,
    embedding_model_name: str,
    embedding_model_size: int,
    context_window_token_limit: int,
    fail_fast: bool,
    openai_api_key: Optional[str],
    anthropic_api_key: Optional[str],
    openai_api_base: Optional[str],
    openai_embedding_api_base: Optional[str],
    openai_organization: Optional[str],
) -> ElroyConfig:
    if chat_model_name in anthropic_models:
        assert anthropic_api_key is not None, "Anthropic API key is required for Anthropic chat models"
        chat_model = ChatModel(
            model=chat_model_name,
            api_key=anthropic_api_key,
            ensure_alternating_roles=True,
        )
    else:
        if chat_model_name in open_ai_chat_completion_models:
            assert openai_api_key is not None, "OpenAI API key is required for OpenAI chat models"
        chat_model = ChatModel(
            model=chat_model_name,
            api_key=openai_api_key,
            ensure_alternating_roles=False,
            api_base=openai_api_base,
            organization=openai_organization,
        )
    if embedding_model_name in open_ai_embedding_models:
        assert openai_api_key is not None, "OpenAI API key is required for OpenAI embedding models"

    embedding_model = EmbeddingModel(
        model=embedding_model_name,
        embedding_size=embedding_model_size,
        api_key=openai_api_key,
        api_base=openai_embedding_api_base,
        organization=openai_organization,
    )

    return ElroyConfig(
        postgres_url=postgres_url,
        chat_model=chat_model,
        embedding_model=embedding_model,
        fail_fast=fail_fast,
        context_window_token_limit=context_window_token_limit,
        context_refresh_token_trigger_limit=int(context_window_token_limit * 0.66),
        context_refresh_token_target=int(context_window_token_limit * 0.33),
        max_in_context_message_age_seconds=int(timedelta(hours=2).total_seconds()),
        context_refresh_interval_seconds=int(timedelta(minutes=10).total_seconds()),
    )


@contextmanager
def session_manager(postgres_url: str) -> Generator[Session, None, None]:
    engine = create_engine(postgres_url, poolclass=NullPool)
    session = Session(engine)
    try:
        yield session
        if session.is_active:  # Only commit if the session is still active
            session.commit()
    except Exception:
        if session.is_active:  # Only rollback if the session is still active
            session.rollback()
        raise
    finally:
        if session.is_active:  # Only close if not already closed
            session.close()


T = TypeVar("T", bound=ElroyIO)


@dataclass
class ElroyContext(Generic[T]):
    session: Session
    io: T
    config: ElroyConfig
    user_id: int



import asyncio
import logging
import os
import sys
from contextlib import contextmanager
from datetime import datetime, timedelta
from typing import Generator, Iterable, Optional

import typer
from colorama import init
from litellm import anthropic_models, open_ai_chat_completion_models
from toolz import concat, pipe, unique
from toolz.curried import filter, map

from ..cli.updater import check_updates, ensure_current_db_migration, version_callback
from ..config.config import ROOT_DIR, ElroyContext, get_config, session_manager
from ..config.constants import (
    CLI_USER_ID,
    DEFAULT_ASSISTANT_COLOR,
    DEFAULT_CHAT_MODEL_NAME,
    DEFAULT_CONTEXT_WINDOW_LIMIT,
    DEFAULT_EMBEDDING_MODEL_NAME,
    DEFAULT_INPUT_COLOR,
    DEFAULT_INTERNAL_THOUGHT_COLOR,
    DEFAULT_SYSTEM_MESSAGE_COLOR,
    DEFAULT_WARNING_COLOR,
    EMBEDDING_SIZE,
    INITIAL_REFRESH_WAIT_SECONDS,
    MIN_CONVO_AGE_FOR_GREETING,
)
from ..docker_postgres import DOCKER_DB_URL, is_docker_running, start_db, stop_db
from ..io.base import StdIO
from ..io.cli import CliIO
from ..logging_config import setup_logging
from ..messaging.context import context_refresh
from ..messaging.messenger import process_message, validate
from ..onboard_user import onboard_user
from ..repository.data_models import SYSTEM, USER, ContextMessage
from ..repository.goals.queries import get_goal_names
from ..repository.memory import get_memory_names, manually_record_user_memory
from ..repository.message import (
    get_context_messages,
    get_time_since_most_recent_user_message,
    replace_context_messages,
)
from ..repository.user import is_user_exists
from ..system_commands import SYSTEM_COMMANDS, contemplate, invoke_system_command
from ..tools.user_preferences import get_user_preferred_name, set_user_preferred_name
from ..utils.clock import get_utc_now
from ..utils.utils import datetime_to_string, run_in_background_thread

app = typer.Typer(help="Elroy CLI", context_settings={"obj": {}})


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version and exit.",
    ),
    postgres_url: Optional[str] = typer.Option(
        None,
        envvar="ELROY_POSTGRES_URL",
        help="Postgres URL to use for Elroy. If set, ovverrides use_docker_postgres.",
    ),
    show_internal_thought_monologue: bool = typer.Option(
        False,
        help="Show the assistant's internal thought monologue like memory consolidation and internal reflection.",
    ),
    openai_api_key: Optional[str] = typer.Option(
        None,
        envvar="OPENAI_API_KEY",
        help="OpenAI API key, required for OpenAI (or OpenAI compatible) models.",
    ),
    openai_api_base: Optional[str] = typer.Option(
        None,
        envvar="OPENAI_API_BASE",
        help="OpenAI API (or OpenAI compatible) base URL.",
    ),
    openai_embedding_api_base: Optional[str] = typer.Option(
        None,
        envvar="OPENAI_API_BASE",
        help="OpenAI API (or OpenAI compatible) base URL for embeddings.",
    ),
    openai_organization: Optional[str] = typer.Option(
        None,
        envvar="OPENAI_ORGANIZATION",
        help="OpenAI (or OpenAI compatible) organization ID.",
    ),
    anthropic_api_key: Optional[str] = typer.Option(
        None,
        envvar="ANTHROPIC_API_KEY",
        help="Anthropic API key, required for Anthropic models.",
    ),
    context_window_token_limit: int = typer.Option(
        DEFAULT_CONTEXT_WINDOW_LIMIT,
        envvar="ELROY_CONTEXT_WINDOW_TOKEN_LIMIT",
        help="How many tokens to keep in context before compressing.",
    ),
    log_file_path: str = typer.Option(
        os.path.join(ROOT_DIR, "logs", "elroy.log"),
        envvar="ELROY_LOG_FILE_PATH",
        help="Where to write logs.",
    ),
    use_docker_postgres: Optional[bool] = typer.Option(
        True,
        envvar="USE_DOCKER_POSTGRES",
        help="If true and postgres_url is not set, will attempt to start a Docker container for Postgres.",
    ),
    stop_docker_postgres_on_exit: Optional[bool] = typer.Option(
        False,
        envvar="STOP_DOCKER_POSTGRES_ON_EXIT",
        help="Whether or not to stop the Postgres container on exit.",
    ),
    system_message_color: str = typer.Option(
        DEFAULT_SYSTEM_MESSAGE_COLOR,
        help="Color for system messages.",
    ),
    user_input_color: str = typer.Option(DEFAULT_INPUT_COLOR, help="Color for user input."),
    assistant_color: str = typer.Option(
        DEFAULT_ASSISTANT_COLOR,
        help="Color for assistant output.",
    ),
    warning_color: str = typer.Option(DEFAULT_WARNING_COLOR, help="Color for warning messages."),
    internal_thought_color: str = typer.Option(
        DEFAULT_INTERNAL_THOUGHT_COLOR,
        help="Color for internal thought messages.",
    ),
    chat_model: str = typer.Option(
        DEFAULT_CHAT_MODEL_NAME,
        help="The model to use for chat completions.",
    ),
    emedding_model: str = typer.Option(
        DEFAULT_EMBEDDING_MODEL_NAME,
        help="The model to use for text embeddings.",
    ),
    embedding_model_size: int = typer.Option(
        EMBEDDING_SIZE,
        help="The size of the embedding model.",
    ),
    fail_fast: bool = typer.Option(
        False,
        help="Whether to fail when errors occur rather than attempting repair. Intended as a debugging aid.",
    ),
):
    """Common parameters."""

    if not postgres_url and not use_docker_postgres:
        raise typer.BadParameter("If postgres_url parameter or ELROY_POSTGRES_URL env var is not set, use_docker_postgres must be True.")

    if postgres_url and use_docker_postgres:
        logging.info("postgres_url is set, ignoring use_docker_postgres set to True")

    ctx.obj = {
        "elroy_config": get_config(
            postgres_url=postgres_url or DOCKER_DB_URL,
            chat_model_name=chat_model,
            fail_fast=fail_fast,
            embedding_model_name=emedding_model,
            embedding_model_size=embedding_model_size,
            context_window_token_limit=context_window_token_limit,
            openai_api_key=openai_api_key,
            anthropic_api_key=anthropic_api_key,
            openai_api_base=openai_api_base,
            openai_embedding_api_base=openai_embedding_api_base,
            openai_organization=openai_organization,
        ),
        "show_internal_thought_monologue": show_internal_thought_monologue,
        "log_file_path": log_file_path,
        "use_docker_postgres": use_docker_postgres,
        "stop_docker_postgres_on_exit": stop_docker_postgres_on_exit,
        "system_message_color": system_message_color,
        "user_input_color": user_input_color,
        "assistant_color": assistant_color,
        "warning_color": warning_color,
        "internal_thought_color": internal_thought_color,
        "is_tty": sys.stdin.isatty(),
    }


@app.command()
def chat(ctx: typer.Context):
    """Start the Elroy chat interface"""

    if not sys.stdin.isatty():
        with init_elroy_context(ctx) as context:
            for line in sys.stdin:
                process_and_deliver_msg(context, line)
        return

    with init_elroy_context(ctx) as context:
        check_updates(context)
        asyncio.run(main_chat(context))
        context.io.sys_message(f"Exiting...")


@app.command()
def remember(
    ctx: typer.Context,
    file: Optional[str] = typer.Option(None, "--file", "-f", help="File to read memory text from"),
):
    """Create a new memory from stdin or interactively"""

    with init_elroy_context(ctx) as context:
        memory_name = None
        if not sys.stdin.isatty():
            memory_text = sys.stdin.read()
            metadata = "Memory ingested from stdin\n" f"Ingested at: {datetime_to_string(datetime.now())}\n"
            memory_text = f"{metadata}\n{memory_text}"
            memory_name = f"Memory from stdin, ingested {datetime_to_string(datetime.now())}"
        elif file:
            try:
                with open(file, "r") as f:
                    memory_text = f.read()
                # Add file metadata
                file_stat = os.stat(file)
                metadata = "Memory ingested from file"
                "File: {file}"
                f"Last modified: {datetime_to_string(datetime.fromtimestamp(file_stat.st_mtime))}\n"
                f"Created at: {datetime_to_string(datetime.fromtimestamp(file_stat.st_ctime))}"
                f"Size: {file_stat.st_size} bytes\n"
                f"Ingested at: {datetime_to_string(datetime.now())}\n"
                memory_text = f"{memory_text}\n{metadata}"
                memory_name = f"Memory from file: {file}, ingested {datetime_to_string(datetime.now())}"
            except Exception as e:
                context.io.sys_message(f"Error reading file: {e}")
                exit(1)
        else:
            # Get the memory text from user
            memory_text = asyncio.run(context.io.prompt_user("Enter the memory text:"))
            memory_text += f"\nManually entered memory, at: {datetime_to_string(datetime.now())}"
            # Optionally get memory name
            memory_name = asyncio.run(context.io.prompt_user("Enter memory name (optional, press enter to skip):"))
        try:
            manually_record_user_memory(context, memory_text, memory_name)
            context.io.sys_message(f"Memory created: {memory_name}")
            exit(0)
        except ValueError as e:
            context.io.assistant_msg(f"Error creating memory: {e}")
            exit(1)


@app.command()
def list_chat_models(ctx: typer.Context):
    """Lists supported chat models"""

    for m in open_ai_chat_completion_models:
        print(f"{m} (OpenAI)")
    for m in anthropic_models:
        print(f"{m} (Anthropic)")


def process_and_deliver_msg(context: ElroyContext, user_input: str, role=USER):
    if user_input.startswith("/") and role == USER:
        cmd = user_input[1:].split()[0]

        if cmd.lower() not in {f.__name__ for f in SYSTEM_COMMANDS}:
            context.io.assistant_msg(f"Unknown command: {cmd}")
        else:
            try:
                context.io.sys_message(invoke_system_command(context, user_input))
            except Exception as e:
                context.io.sys_message(f"Error invoking system command: {e}")
    else:
        context.io.assistant_msg(process_message(context, user_input, role))


def periodic_context_refresh(context: ElroyContext, interval_seconds: float):
    """Run context refresh in a background thread"""
    # Create new event loop for this thread
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    async def refresh_loop(context: ElroyContext):
        await asyncio.sleep(INITIAL_REFRESH_WAIT_SECONDS)
        while True:
            try:
                logging.info("Refreshing context")
                await context_refresh(context)  # Keep this async
                await asyncio.sleep(interval_seconds)
            except Exception as e:
                logging.error(f"Error in periodic context refresh: {e}")
                context.session.rollback()

    try:
        # hack to get a new session for the thread
        with session_manager(context.config.postgres_url) as session:
            loop.run_until_complete(
                refresh_loop(
                    ElroyContext(
                        user_id=CLI_USER_ID,
                        session=session,
                        config=context.config,
                        io=context.io,
                    )
                )
            )
    finally:
        loop.close()


async def main_chat(context: ElroyContext[CliIO]):
    init(autoreset=True)

    run_in_background_thread(
        periodic_context_refresh,
        context,
        context.config.context_refresh_interval_seconds,
    )

    context.io.print_title_ruler()

    if not is_user_exists(context):
        context.io.notify_warning("Elroy is in alpha release")
        name = await context.io.prompt_user("Welcome to Elroy! What should I call you?")
        user_id = onboard_user(context.session, context.io, context.config, name)
        assert isinstance(user_id, int)

        set_user_preferred_name(context, name)
        _print_memory_panel(context, get_context_messages(context))
        process_and_deliver_msg(context, "Elroy user {name} has been onboarded. Say hello and introduce yourself.", role=SYSTEM)

    else:
        context_messages = get_context_messages(context)

        validated_messages = validate(context.config, context_messages)

        if context_messages != validated_messages:
            replace_context_messages(context, validated_messages)
            logging.warning("Context messages were repaired")
            context_messages = get_context_messages(context)

        _print_memory_panel(context, context_messages)

        if (get_time_since_most_recent_user_message(context_messages) or timedelta()) < MIN_CONVO_AGE_FOR_GREETING:
            logging.info("User has interacted recently, skipping greeting.")

        else:
            preferred_name = get_user_preferred_name(context)

            # TODO: should include some information about how long the user has been talking to Elroy
            process_and_deliver_msg(
                context,
                f"{preferred_name} has logged in. The current time is {datetime_to_string(datetime.now())}. I should offer a brief greeting.",
                SYSTEM,
            )

    while True:
        try:

            context.io.update_completer(get_goal_names(context), get_memory_names(context))

            user_input = await context.io.prompt_user()
            if user_input.lower().startswith("/exit") or user_input == "exit":
                break
            elif user_input:
                process_and_deliver_msg(context, user_input)
                run_in_background_thread(contemplate, context)
        except EOFError:
            break

        context.io.rule()
        _print_memory_panel(context, get_context_messages(context))


@contextmanager
def init_elroy_context(ctx: typer.Context) -> Generator[ElroyContext, None, None]:
    """Create an ElroyContext as a context manager"""

    if ctx.obj["is_tty"]:
        io = CliIO(
            ctx.obj["show_internal_thought_monologue"],
            ctx.obj["system_message_color"],
            ctx.obj["assistant_color"],
            ctx.obj["user_input_color"],
            ctx.obj["warning_color"],
            ctx.obj["internal_thought_color"],
        )
    else:
        io = StdIO()

    try:
        setup_logging(ctx.obj["log_file_path"])

        if ctx.obj["use_docker_postgres"]:
            if is_docker_running():
                start_db()
            else:
                raise typer.BadParameter(
                    "Elroy was started with use_docker_postgres set to True, but no Docker container is running. Please either start a Docker container, provide a postgres_url parameter, or set the ELROY_POSTGRES_URL environment variable."
                )

        # Check if migrations need to be run
        config = ctx.obj["elroy_config"]
        ensure_current_db_migration(io, config.postgres_url)

        with session_manager(config.postgres_url) as session:
            yield ElroyContext(
                user_id=CLI_USER_ID,
                session=session,
                config=config,
                io=io,
            )

    finally:
        if ctx.obj["use_docker_postgres"] and ctx.obj["stop_docker_postgres_on_exit"]:
            io.sys_message("Stopping Docker Postgres container...")
            stop_db()


def _print_memory_panel(context: ElroyContext, context_messages: Iterable[ContextMessage]) -> None:
    pipe(
        context_messages,
        filter(
            lambda m: not m.created_at
            or m.created_at > get_utc_now() - timedelta(seconds=context.config.max_in_context_message_age_seconds)
        ),
        map(lambda m: m.memory_metadata),
        filter(lambda m: m is not None),
        concat,
        map(lambda m: f"{m.memory_type}: {m.name}"),
        unique,
        list,
        sorted,
        context.io.print_memory_panel,
    )


def main():
    if len(sys.argv) == 1:
        sys.argv.append("chat")
    app()


if __name__ == "__main__":
    main()



import logging
from functools import partial
from typing import Dict, Iterator, List, NamedTuple, Optional, Union

from litellm.types.utils import ChatCompletionDeltaToolCall
from toolz import juxt, pipe
from toolz.curried import do, filter, map, remove, tail

from ..config.config import ChatModel, ElroyConfig, ElroyContext
from ..config.constants import (
    SYSTEM_INSTRUCTION_LABEL,
    MisplacedSystemInstructError,
    MissingAssistantToolCallError,
    MissingSystemInstructError,
    MissingToolCallMessageError,
)
from ..llm.client import generate_chat_completion_message, get_embedding
from ..repository.data_models import ASSISTANT, SYSTEM, TOOL, USER
from ..repository.embeddings import get_most_relevant_goal, get_most_relevant_memory
from ..repository.facts import to_fact
from ..repository.message import (
    ContextMessage,
    MemoryMetadata,
    get_context_messages,
    is_system_instruction,
    replace_context_messages,
)
from ..tools.function_caller import FunctionCall, PartialToolCall, exec_function_call
from ..utils.utils import last_or_none, logged_exec_time


class ToolCallAccumulator:
    def __init__(self, chat_model: ChatModel):
        self.chat_model = chat_model
        self.tool_calls: Dict[int, PartialToolCall] = {}
        self.last_updated_index: Optional[int] = None

    def update(self, delta_tool_calls: Optional[List[ChatCompletionDeltaToolCall]]) -> Iterator[FunctionCall]:
        for delta in delta_tool_calls or []:
            if delta.index not in self.tool_calls:
                if (
                    self.last_updated_index is not None
                    and self.last_updated_index in self.tool_calls
                    and self.last_updated_index != delta.index
                ):
                    raise ValueError("New tool call started, but old one is not yet complete")
                assert delta.id
                self.tool_calls[delta.index] = PartialToolCall(id=delta.id, model=self.chat_model.model)

            completed_tool_call = self.tool_calls[delta.index].update(delta)
            if completed_tool_call:
                self.tool_calls.pop(delta.index)
                yield completed_tool_call
            else:
                self.last_updated_index = delta.index


def process_message(context: ElroyContext, msg: str, role: str = USER) -> Iterator[str]:
    assert role in [USER, ASSISTANT, SYSTEM]

    context_messages = pipe(
        get_context_messages(context),
        partial(validate, context.config),
        list,
        lambda x: x + [ContextMessage(role=role, content=msg, chat_model=None)],
        lambda x: x + get_relevant_memories(context, x),
        list,
    )

    full_content = ""

    while True:
        function_calls: List[FunctionCall] = []
        tool_context_messages: List[ContextMessage] = []

        for stream_chunk in _generate_assistant_reply(context.config.chat_model, context_messages):
            if isinstance(stream_chunk, ContentItem):
                full_content += stream_chunk.content
                yield stream_chunk.content
            elif isinstance(stream_chunk, FunctionCall):
                pipe(
                    stream_chunk,
                    do(function_calls.append),
                    lambda x: ContextMessage(
                        role=TOOL,
                        tool_call_id=x.id,
                        content=exec_function_call(context, x),
                        chat_model=context.config.chat_model.model,
                    ),
                    tool_context_messages.append,
                )
        context_messages.append(
            ContextMessage(
                role=ASSISTANT,
                content=full_content,
                tool_calls=(None if not function_calls else [f.to_tool_call() for f in function_calls]),
                chat_model=context.config.chat_model.model,
            )
        )

        if not tool_context_messages:
            replace_context_messages(context, context_messages)
            break
        else:
            context_messages += tool_context_messages


def validate(config: ElroyConfig, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    return pipe(
        context_messages,
        partial(_validate_system_instruction_correctly_placed, config.fail_fast),
        partial(_validate_assistant_tool_calls_followed_by_tool, config.fail_fast),
        partial(_validate_tool_messages_have_assistant_tool_call, config.fail_fast),
        lambda msgs: (
            msgs if not config.chat_model.ensure_alternating_roles else validate_first_user_precedes_first_assistant(config.fail_fast, msgs)
        ),
        list,
    )


def validate_first_user_precedes_first_assistant(fail_fast: bool, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    user_and_assistant_messages = [m for m in context_messages if m.role in [USER, ASSISTANT]]

    if user_and_assistant_messages and user_and_assistant_messages[0].role != USER:
        if fail_fast:
            raise ValueError("First non-system message must be USER role for this model")
        else:
            context_messages = [
                context_messages[0],
                ContextMessage(role=USER, content="The user has begun the converstaion", chat_model=None),
            ] + context_messages[1:]
    return context_messages


def _validate_system_instruction_correctly_placed(fail_fast: bool, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    validated_messages = []

    for idx, message in enumerate(context_messages):
        if idx == 0 and not is_system_instruction(message):
            if fail_fast:
                raise MissingSystemInstructError()
            else:
                logging.error(f"First message is not system instruction, repairing by inserting system instruction")
                validated_messages += [
                    ContextMessage(
                        role=SYSTEM, content=f"{SYSTEM_INSTRUCTION_LABEL}\nYou are Elroy, a helpful assistant.", chat_model=None
                    ),
                    message,
                ]
        elif idx != 0 and is_system_instruction(message):
            if fail_fast:
                raise MisplacedSystemInstructError()
            else:
                logging.error("Found system message in non-first position, repairing by dropping message")
                continue
        else:
            validated_messages.append(message)
    return validated_messages


def _validate_assistant_tool_calls_followed_by_tool(fail_fast: bool, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    """
    Validates that any assistant message with non-empty tool_calls is followed by corresponding tool messages.
    """

    for idx, message in enumerate(context_messages):
        if (message.role == ASSISTANT and message.tool_calls is not None) and (
            idx == len(context_messages) - 1 or context_messages[idx + 1].role != TOOL
        ):
            if fail_fast:
                raise MissingToolCallMessageError()
            else:
                logging.error(
                    f"Assistant message with tool_calls not followed by tool message: ID = {message.id}, repairing by removing tool_calls"
                )
                message.tool_calls = None
    return context_messages


def _validate_tool_messages_have_assistant_tool_call(fail_fast: bool, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    """
    Validates that all tool messages have a preceding assistant message with the corresponding tool_calls.
    """

    validated_context_messages = []
    for idx, message in enumerate(context_messages):
        if message.role == TOOL and not _has_assistant_tool_call(message.tool_call_id, context_messages[:idx]):
            if fail_fast:
                raise MissingAssistantToolCallError(f"Message id: {message.id}")
            else:
                logging.warning(
                    f"Tool message without preceding assistant message with tool_calls: ID = {message.id}. Repairing by removing tool message"
                )
                continue
        else:
            validated_context_messages.append(message)

    return validated_context_messages


def _has_assistant_tool_call(tool_call_id: Optional[str], context_messages: List[ContextMessage]) -> bool:
    """
    Assistant tool call message must be in the most recent assistant message
    """
    if not tool_call_id:
        logging.warning("Tool call ID is None")
        return False

    return pipe(
        context_messages,
        filter(lambda x: x.role == ASSISTANT),
        last_or_none,
        lambda msg: msg.tool_calls or [] if msg else [],
        map(lambda x: x.id),
        filter(lambda x: x == tool_call_id),
        any,
    )


@logged_exec_time
def get_relevant_memories(context: ElroyContext, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    from .context import is_memory_in_context

    message_content = pipe(
        context_messages,
        remove(lambda x: x.role == SYSTEM),
        tail(4),
        map(lambda x: f"{x.role}: {x.content}" if x.content else None),
        remove(lambda x: x is None),
        list,
        "\n".join,
    )

    if not message_content:
        return []

    assert isinstance(message_content, str)

    new_memory_messages = pipe(
        message_content,
        partial(get_embedding, context.config.embedding_model),
        lambda x: juxt(get_most_relevant_goal, get_most_relevant_memory)(context, x),
        filter(lambda x: x is not None),
        remove(partial(is_memory_in_context, context_messages)),
        map(
            lambda x: ContextMessage(
                role=SYSTEM,
                memory_metadata=[MemoryMetadata(memory_type=x.__class__.__name__, id=x.id, name=x.get_name())],
                content="Information recalled from assistant memory: " + to_fact(x),
                chat_model=None,
            )
        ),
        list,
    )

    return new_memory_messages


from typing import Iterator


class ContentItem(NamedTuple):
    content: str


StreamItem = Union[ContentItem, FunctionCall]


def _generate_assistant_reply(
    chat_model: ChatModel,
    context_messages: List[ContextMessage],
) -> Iterator[StreamItem]:

    if context_messages[-1].role == ASSISTANT:
        raise ValueError("Assistant message already the most recent message")

    tool_call_accumulator = ToolCallAccumulator(chat_model)
    for chunk in generate_chat_completion_message(chat_model, context_messages):
        if chunk.choices[0].delta.content:  # type: ignore
            yield ContentItem(content=chunk.choices[0].delta.content)  # type: ignore
        if chunk.choices[0].delta.tool_calls:  # type: ignore
            yield from tool_call_accumulator.update(chunk.choices[0].delta.tool_calls)  # type: ignore



"""optional goal cols

Revision ID: 1138d225c857
Revises: 1bf9bd811903
Create Date: 2024-11-01 15:09:28.123613

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1138d225c857"
down_revision: Union[str, None] = "1bf9bd811903"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("goal", "description", existing_type=sa.VARCHAR(), nullable=True)
    op.alter_column("goal", "strategy", existing_type=sa.VARCHAR(), nullable=True)
    op.alter_column("goal", "end_condition", existing_type=sa.VARCHAR(), nullable=True)
    op.drop_column("userpreference", "user_time_zone")
    op.drop_column("userpreference", "display_internal_monologue")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("userpreference", sa.Column("display_internal_monologue", sa.BOOLEAN(), autoincrement=False, nullable=True))
    op.add_column("userpreference", sa.Column("user_time_zone", sa.VARCHAR(), autoincrement=False, nullable=True))
    op.drop_index(op.f("ix_memory_id"), table_name="memory")
    op.create_index("ix_archivalmemory_id", "memory", ["id"], unique=False)
    op.alter_column("goal", "end_condition", existing_type=sa.VARCHAR(), nullable=False)
    op.alter_column("goal", "strategy", existing_type=sa.VARCHAR(), nullable=False)
    op.alter_column("goal", "description", existing_type=sa.VARCHAR(), nullable=False)
    # ### end Alembic commands ###



from ..repository.data_models import EmbeddableSqlModel, Goal, Memory


def to_fact(row: EmbeddableSqlModel) -> str:
    if isinstance(row, Goal):
        from ..repository.goals.operations import goal_to_fact

        return goal_to_fact(row)
    elif isinstance(row, Memory):
        from ..repository.memory import memory_to_fact

        return memory_to_fact(row)
    else:
        raise ValueError(f"Unsupported type: {type(row)}")



import logging
import threading
import time
from datetime import datetime
from functools import partial
from typing import Iterator, Optional, TypeVar

from pytz import UTC

from ..config.config import ElroyContext, session_manager

T = TypeVar("T")


def logged_exec_time(func, name: Optional[str] = None):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        elapsed_time = time.time() - start_time

        if name:
            func_name = name
        else:
            func_name = func.__name__ if not isinstance(func, partial) else func.func.__name__

        logging.info(f"Function '{func_name}' executed in {elapsed_time:.4f} seconds")
        return result

    return wrapper


def first_or_none(iterable: Iterator[T]) -> Optional[T]:
    return next(iterable, None)


def last_or_none(iterable: Iterator[T]) -> Optional[T]:
    return next(reversed(list(iterable)), None)


def datetime_to_string(dt: Optional[datetime]) -> Optional[str]:
    if dt:
        return dt.strftime("%A, %B %d, %Y %I:%M %p %Z")


utc_epoch_to_datetime_string = lambda epoch: datetime_to_string(datetime.fromtimestamp(epoch, UTC))


def run_in_background_thread(fn, context, *args):
    # hack to get a new session for the thread
    with session_manager(context.config.postgres_url) as session:
        thread = threading.Thread(
            target=fn,
            args=(ElroyContext(user_id=context.user_id, session=session, config=context.config, io=context.io), *args),
            daemon=True,
        )
        thread.start()



# Should have param for checking if a similar goal already exists
import logging
from typing import Optional

from sqlmodel import select, update
from toolz import pipe
from toolz.curried import filter

from ...config.config import ElroyContext
from ...messaging.context import (
    drop_goal_from_current_context_only,
    remove_from_context,
)
from ...utils.clock import get_utc_now, string_to_timedelta
from ..data_models import SYSTEM, ContextMessage, Goal
from ..embeddings import upsert_embedding
from ..message import add_context_messages


def goal_to_fact(goal: Goal) -> str:

    return pipe(
        [
            f"# {goal.__class__.__name__}: {goal.name}",
            goal.description if goal.description else None,
            f"## Strategy\n{goal.strategy}" if goal.strategy else None,
            f"## End Condition\n{goal.end_condition}" if goal.end_condition else None,
            f"## Target Completion Time\n{goal.target_completion_time}" if goal.target_completion_time else None,
            "## Status Updates\n" + ("\n".join(goal.status_updates) if goal.status_updates else "No status updates"),
            f"## Priority\n{goal.priority}" if goal.priority else None,
            f"### Note for assistant:\nInformation about this goal should be kept up to date via AI assistant functions: {add_goal_status_update.__name__}, and {mark_goal_completed.__name__}",
        ],
        filter(lambda x: x is not None),
        "\n\n".join,
    )  # type: ignore


def create_goal(
    context: ElroyContext,
    goal_name: str,
    strategy: Optional[str] = None,
    description: Optional[str] = None,
    end_condition: Optional[str] = None,
    time_to_completion: Optional[str] = None,
    priority: Optional[int] = None,
) -> str:
    """Creates a goal. The goal can be for the AI user, or for the assistant in relation to helping the user somehow.
    Goals should be *specific* and *measurable*. They should be based on the user's needs and desires, and should
    be achievable within a reasonable timeframe.

    Args:
        session (Session): The database session.
        user_id (int): user id
        goal_name (str): Name of the goal
        strategy (str): The strategy to achieve the goal. Your strategy should detail either how you (the personal assistant) will achieve the goal, or how you will assist your user to solve the goal. Limit to 100 words.
        description (str): A brief description of the goal. Limit to 100 words.
        end_condition (str): The condition that indicate to you (the personal assistant) that the goal is achieved or terminated. It is critical that this end condition be OBSERVABLE BY YOU (the assistant). For example, the end_condition may be that you've asked the user about the goal status.
        time_to_completion (str): The amount of time from now until the goal can be completed. Should be in the form of NUMBER TIME_UNIT, where TIME_UNIT is one of HOURS, DAYS, WEEKS, MONTHS. For example, "1 DAYS" would be a goal that should be completed within 1 day.
        priority (int): The priority of the goal, from 0-4. Priority 0 is the highest priority, and 4 is the lowest.
    """
    existing_goal = context.session.exec(
        select(Goal).where(
            Goal.user_id == context.user_id,
            Goal.name == goal_name,
            Goal.is_active == True,
        )
    ).one_or_none()
    if existing_goal:
        raise Exception(f"Active goal {goal_name} already exists for user {context.user_id}")
    else:
        goal = Goal(
            user_id=context.user_id,
            name=goal_name,
            description=description,
            strategy=strategy,
            end_condition=end_condition,
            priority=priority,
            target_completion_time=get_utc_now() + string_to_timedelta(time_to_completion) if time_to_completion else None,
        )
        context.session.add(goal)
        context.session.commit()
        context.session.refresh(goal)

        add_context_messages(
            context,
            [
                ContextMessage(
                    role=SYSTEM,
                    content=f"New goal created: {goal_to_fact(goal)}",
                    memory_metadata=[goal.to_memory_metadata()],
                    chat_model=context.config.chat_model.model,
                )
            ],
        )

        upsert_embedding(context, goal)

        return f"Goal '{goal_name}' has been created."


def rename_goal(context: ElroyContext, old_goal_name: str, new_goal_name: str) -> None:
    """Renames an existing active goal.

    Args:
        context (ElroyContext): The Elroy context.
        old_goal_name (str): The current name of the goal.
        new_goal_name (str): The new name for the goal.

    Raises:
        Exception: If the goal with old_goal_name doesn't exist or if a goal with new_goal_name already exists.
    """
    # Check if the old goal exists and is active
    old_goal = context.session.exec(
        select(Goal).where(
            Goal.user_id == context.user_id,
            Goal.name == old_goal_name,
            Goal.is_active == True,
        )
    ).first()
    if not old_goal:
        raise Exception(f"Active goal '{old_goal_name}' not found for user {context.user_id}")

    # Check if a goal with the new name already exists
    existing_goal = context.session.exec(
        select(Goal).where(
            Goal.user_id == context.user_id,
            Goal.name == new_goal_name,
            Goal.is_active == True,
        )
    ).first()
    if existing_goal:
        raise Exception(f"Active goal '{new_goal_name}' already exists for user {context.user_id}")

    # we need to drop the goal from context as the metadata includes the goal name.
    drop_goal_from_current_context_only(context, old_goal.name)

    # Rename the goal
    old_goal.name = new_goal_name
    old_goal.updated_at = get_utc_now()

    context.session.commit()
    context.session.refresh(old_goal)

    upsert_embedding(context, old_goal)

    add_context_messages(
        context,
        [
            ContextMessage(
                role=SYSTEM,
                content=f"Goal '{old_goal_name}' has been renamed to '{new_goal_name}': {goal_to_fact(old_goal)}",
                memory_metadata=[old_goal.to_memory_metadata()],
                chat_model=context.config.chat_model.model,
            )
        ],
    )


def _update_goal_status(context: ElroyContext, goal_name: str, is_terminal: bool, status: Optional[str]) -> None:
    goal = context.session.exec(
        select(Goal).where(
            Goal.user_id == context.user_id,
            Goal.name == goal_name,
            Goal.is_active == True,
        )
    ).first()
    if not goal:
        raise Exception(f"Active goal {goal_name} not found for user {context.user_id}")

    logging.info(f"Updating goal {goal_name} for user {context.user_id}")
    logging.info(f"Current status updates: {goal.status_updates}")

    # Append the new status update to the list
    if status:
        if goal.status_updates is None:
            goal.status_updates = []
        goal.status_updates.append(status)

    logging.info(f"Updated status updates: {goal.status_updates}")

    # Update the goal's active status if it's terminal
    if is_terminal:
        goal.is_active = False

    goal.updated_at = get_utc_now()

    # Explicitly update the status_updates column, the recommended style has a bug
    context.session.execute(
        update(Goal)
        .where(Goal.id == goal.id)  # type: ignore
        .values(status_updates=goal.status_updates, is_active=goal.is_active, updated_at=goal.updated_at)
    )

    context.session.commit()

    assert status in goal.status_updates, "Status update not found in goal status updates"

    # Refresh the goal object after commit
    context.session.refresh(goal)

    logging.info(f"Status updates after commit and refresh: {goal.status_updates}")

    assert goal.id

    upsert_embedding(context, goal)

    if not goal.is_active:

        remove_from_context(context, goal)


def add_goal_status_update(context: ElroyContext, goal_name: str, status_update_or_note: str) -> str:
    """Captures either a progress update or note relevant to the goal.

    Args:
        session (Session): The database session.
        user_id (int): The user id
        goal_name (str): Name of the goal
        status_update_or_note (str): A brief status update or note about either progress or learnings relevant to the goal. Limit to 100 words.
    Returns:
        str: Confirmation message
    """
    logging.info(f"Updating goal {goal_name} for user {context.user_id}")
    _update_goal_status(context, goal_name, False, status_update_or_note)

    return f"Status update added to goal '{goal_name}'."


def create_onboarding_goal(context: ElroyContext, preferred_name: str) -> None:

    create_goal(
        context=context,
        goal_name=f"Introduce myself to {preferred_name}",
        description="Introduce myself - a few things that make me unique are my ability to form long term memories, and the ability to set and track goals.",
        strategy=f"After exchanging some pleasantries, tell {preferred_name} about my ability to form long term memories, including goals. Use function {add_goal_status_update.__name__} with any progress or learnings.",
        end_condition=f"{preferred_name} has been informed about my ability to track goals",
        priority=1,
        time_to_completion="1 HOUR",
    )


def mark_goal_completed(context: ElroyContext, goal_name: str, closing_comments: Optional[str] = None) -> str:
    """Marks a goal as completed, with closing comments.

    Args:
        session (Session): The database session.
        user_id (int): The user ID
        goal_name (str): The name of the goal
        closing_comments (str): Updated status with a short account of how the goal was completed and what was learned.
    Returns:
        str: Confirmation message
    """
    _update_goal_status(
        context,
        goal_name,
        True,
        closing_comments,
    )

    return f"Goal '{goal_name}' has been marked as completed."


def delete_goal_permamently(context: ElroyContext, goal_name: str) -> str:
    """Closes the goal.

    Args:
        session (Session): The database session.
        user_id (int): The user ID
        goal_name (str): The name of the goal
    Returns:
        str: Result of the deletion
    """

    _update_goal_status(
        context,
        goal_name,
        True,
        "Goal has been deleted",
    )

    return f"Goal '{goal_name}' has been deleted."



from functools import partial
from typing import Optional, Tuple

from ..config.config import ChatModel
from ..config.constants import INNER_THOUGHT_TAG, MEMORY_WORD_COUNT_LIMIT
from ..llm.client import query_llm_json, query_llm_with_word_limit

ONBOARDING_SYSTEM_SUPPLEMENT_INSTRUCT = (
    lambda preferred_name: f"""
This is the first exchange between you and your primary user, {preferred_name}.

Greet {preferred_name} warmly and introduce yourself.

In these early messages, prioritize learning some basic information about {preferred_name}.

However, avoid asking too many questions at once. Be sure to engage in a natural conversation. {preferred_name} is likely unsure of what to expect from you, so be patient and understanding.
"""
)

summarize_conversation = partial(
    query_llm_with_word_limit,
    word_limit=MEMORY_WORD_COUNT_LIMIT,
    system="""
Your job is to summarize a history of previous messages in a conversation between an AI persona and a human.
The conversation you are given is a from a fixed context window and may not be complete.
Messages sent by the AI are marked with the 'assistant' role.
Summarize what happened in the conversation from the perspective of ELROY (use the first person).
Note not only the content of the messages but also the context and the relationship between the entities mentioned.
Also take note of the overall tone of the conversation. For example, the user might be engaging in terse question and answer, or might be more conversational.
Only output the summary, do NOT include anything else in your output.
""",
)


async def summarize_for_memory(model: ChatModel, user_preferred_name: str, conversation_summary: str) -> Tuple[str, str]:
    response = query_llm_json(
        model=model,
        prompt=conversation_summary,
        system=f"""
You are the internal thought monologue of an AI personal assistant, forming a memory from a conversation.

Given a conversation summary, your will reflect on the conversation and decide which memories might be relevant in future interactions with {user_preferred_name}.

Pay particular attention facts about {user_preferred_name}, such as name, age, location, etc.
Specifics about events and dates are also important.

When referring to dates and times, use use ISO 8601 format, rather than relative references.
If an event is recurring, specify the frequency, start datetime, and end datetime if applicable.

Focus on facts in the real world, as opposed to facts about the conversation itself. However, it is also appropriate to draw conclusions from the infromation in the conversation.

Your response should be in the voice of an internal thought monolgoue, and should be understood to be as part of an ongoing conversation.

Don't say things like "finally, we talked about", or "in conclusion", as this is not the end of the conversation.

Return your response in JSON format, with the following structure:
- TITLE: the title of the archival memory
- {INNER_THOUGHT_TAG}: the internal thought monologue
""",
    )

    return (response["TITLE"], response[INNER_THOUGHT_TAG])  # type: ignore


DEFAULT_CONTEMPLATE_PROMPT = "Think about the conversation you're in the middle of having. What are important facts to remember?"
"What conclusions can you draw?"
"Also consider if any functions might be appropriate to invoke, and why"


def contemplate_prompt(user_preferred_name: str, prompt: Optional[str]) -> str:
    prompt = prompt or DEFAULT_CONTEMPLATE_PROMPT

    return f"""
You are the internal thought monologue of an AI personal assistant, forming a memory from a conversation.

Given a conversation summary with your user, {user_preferred_name}, your will reflect on a prompt.

In your reflection, focus on the more recent messages within the conversation.

Your prompt is:

{prompt}

Style guidance:
- Above all be concise, your response should be no more than 30 words, or 1-3 sentences.
- If you refer to dates and times, use ISO 8601 format, rather than relative references.
- Your response should be in the first person voice of the assistant internal thought monolgoue, and should be understood to be as part of an ongoing conversation.
- "Don't say things like 'finally, we talked about', or 'in conclusion', as this is not the end of the conversation."
"""
